package com.keyware.composeanalysis.task;


import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.lang.Pair;
import com.alibaba.fastjson.JSONArray;
import com.keyware.common.constant.enums.AnalysisStatusEnum;
import com.keyware.common.exception.BusinessException;
import com.keyware.composeanalysis.constant.FixedValue;
import com.keyware.composeanalysis.constant.RedisConst;
import com.keyware.composeanalysis.constant.SolrDBConst;
import com.keyware.composeanalysis.constant.enums.AnalysisLevelEnum;
import com.keyware.composeanalysis.constant.enums.FileAnalysisStatusEnum;
import com.keyware.composeanalysis.entity.AnalysisTask;
import com.keyware.composeanalysis.mongo.FileDataMongoDto;
import com.keyware.composeanalysis.mongo.LineDataMongoDto;
import com.keyware.composeanalysis.mongo.MatchOpenFile;
import com.keyware.composeanalysis.mongo.MatchOpenFileMongoDto;
import com.keyware.composeanalysis.solr.FunctionInfo;
import com.keyware.composeanalysis.solr.VersionTree;
import com.keyware.composeanalysis.util.*;
import com.keyware.keyswan.common.LineModel;
import com.keyware.keyware.anaysis.Analysis;
import com.keyware.keyware.anaysis.AnalysisFactory;
import com.keyware.keyware.common.CodeFile;
import com.keyware.keyware.common.Function;
import com.keyware.utils.IdGenerator;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrDocument;
import org.apache.solr.common.SolrDocumentList;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Update;

import java.io.FileInputStream;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.springframework.data.mongodb.core.query.Criteria.where;

/**
 * @author liuzongren
 * @ClassName LineAnalysisTask
 * @description: 函数级别溯源 任务
 * @datetime 2024年 07月 25日 16:19
 * @version: 1.0
 */

@Log4j2
public class FunctionAnalysisTask extends IAnalysisTask {

    private MongoTemplate mongoTemplate;
    private AnalysisTask analysisTask;
    //被测件的文件信息
    private FileDataMongoDto analysisFile;

    private SolrUtils solrUtils;

    private RedisUtil redisUtil;

    private CountDownLatch countDownLatch;


    public FunctionAnalysisTask(AnalysisTask analysisTask, FileDataMongoDto analysisFile, MongoTemplate mongoTemplate, CountDownLatch countDownLatch) {
        this.mongoTemplate = mongoTemplate;
        this.analysisTask = analysisTask;
        this.analysisFile = analysisFile;
        this.countDownLatch = countDownLatch;
        this.solrUtils = SpringContextUtils.getBean(SolrUtils.class);
        this.redisUtil = SpringContextUtils.getBean(RedisUtil.class);
    }

    /**
     * 方法 或者代码块 级别  源代码溯源
     * 当前任务 需要在 文件级分析完成后 进行
     */

    @Override
    public void run() {
        //执行任务前,判断一下任务执行的状态
        Object status = redisUtil.get(String.format(RedisConst.TASK_RUNNING_STATUS_KEY_PREFIX, analysisTask.getId()));
        if (status != null && (status.equals(AnalysisStatusEnum.STOP_ANALYSIS.getCode()) || status.equals(AnalysisStatusEnum.PAUSE_ANALYSIS.getCode()))) {
            log.info("任务已取消,fileName:{}", analysisFile.getName());
            countDownLatch.countDown();
            return;
        }
        //获取文件地址
        String filePath = analysisFile.getFileUrl();
        //获取文件名称
        String fileName = analysisFile.getName();

        try {

            //根据文件后缀判断需要查询的solr特征库库名称
            String featureCoreName = FixedValue.SUFFIX_SOLR_FILE.get(analysisFile.getSuffix());

            //根据文件后缀,去检索sourceFileBase库,来获取文件版本信息
            String sourceFileBaseCoreName = FixedValue.SUFFIX_SOLR_VERSION.get(analysisFile.getSuffix());

            //根据文件的名称获取函数解析器
            Analysis analysis = AnalysisFactory.getAnalysis(fileName);

            if (analysis == null){
                throw  new BusinessException("获取文件解析器失败,文件名称："+fileName);
            }

            //解析文件
            CodeFile codeFile = analysis.analysisFile(new FileInputStream(filePath));

            //根据函数特征去匹配到开源文件
            SolrDocumentList matchOpenFiles = getFeatureSimilarityFromSolr(featureCoreName,  codeFile.getFunctionList());

            //计算开源率
            doAnalysis(matchOpenFiles, sourceFileBaseCoreName, codeFile);

            //更新文件表的分析状态为3 函数级特征以分析完毕
            analysisFile.setFileAnalysisStatus(FileAnalysisStatusEnum.ANALYSIS_DONE.getCode());
            mongoTemplate.update(FileDataMongoDto.class)
                    .matching(where("_id").is(analysisFile.getId()))
                    .replaceWith(analysisFile)
                    .findAndReplace();

            AnalysisLogUtil.insert(mongoTemplate, "【函数级分析】完成" + fileName);
            log.info("文件" + fileName + ":函数级分析完成");
        } catch (Exception e) {
            AnalysisLogUtil.insertErrorInfo(mongoTemplate, "【函数级级分析】失败" + fileName, e);
            log.error("文件:" + fileName + "【函数级级分析】失败!", e);
            //修改当前文件分析状态未失败
            mongoTemplate.update(FileDataMongoDto.class)
                    .matching(where("_id").is(analysisFile.getId()))
                    .apply(new Update().set("fileAnalysisStatus", FileAnalysisStatusEnum.FAILED_ANALYSIS.getCode()))
                    .first();
        } finally {
            countDownLatch.countDown();
        }
    }


    /**
     * 对比函数级文本相似度
     *
     * @param matchOpenFiles         通过特征匹配到的开源文件的md5
     * @param sourceFileBaseCoreName 查询版开源文件版本ID的 solr库名称
     * @param fileAnalysisRes        被测件的函数解析结果
     */
    private void doAnalysis(SolrDocumentList matchOpenFiles, String sourceFileBaseCoreName, CodeFile fileAnalysisRes) {

        if (CollectionUtil.isEmpty(matchOpenFiles)) {
            return;
        }

        //按照函数的特征md5进行分组,getter ,setter等方法的 特征值会重复
        Map<String, List<Function>> featureMd5FunctionMap = fileAnalysisRes.getFunctionList().stream().collect(Collectors.groupingBy(Function::getMd5));

        //函数代码总函数
        BigDecimal totalFunctionLineCount = new BigDecimal(fileAnalysisRes.getFunctionList().stream().mapToInt(Function::getCodeRowNum).sum());

        //匹配到的特征函数Md5
        Set<String> matchFeatureFunctionMd5s = new HashSet();

        //匹配到源码的行号
        Set<Integer> matchOpenLineRowsNum = new HashSet();

        //计算与每个开源文件的开源率和特征相似度
        List<MatchOpenFile> matchOpenFilesRes = calculateSimilarityAndOpenRate(matchOpenFiles, fileAnalysisRes, sourceFileBaseCoreName, matchOpenLineRowsNum, matchFeatureFunctionMd5s);

        //计算文件的总体的特征相似度
        int matchFunctionLineCount = 0;
        for (String matchFeatureFunctionMd5 : matchFeatureFunctionMd5s) {
            matchFunctionLineCount += featureMd5FunctionMap.get(matchFeatureFunctionMd5).stream().mapToInt(Function::getCodeRowNum).sum();
        }

        BigDecimal featureSimilarity = new BigDecimal(matchFunctionLineCount).divide(totalFunctionLineCount, 4, RoundingMode.HALF_UP).multiply(new BigDecimal(100)).setScale(2);

        //计算文件的总体开源率
        BigDecimal openRate = new BigDecimal(matchOpenLineRowsNum.size()).divide(new BigDecimal(analysisFile.getCodeRowNum()), 4, RoundingMode.HALF_UP).multiply(new BigDecimal(100)).setScale(2);

        //获取开源率的阈值
        Integer openRateThreshold = analysisTask.getOpenRateThreshold();

        //如果开源率大于阈值，则将当前文件设置成开源
        if (openRate.floatValue() > openRateThreshold) {
            analysisFile.setOpenType(true);
        }

        //保存当前文件开源行数
        analysisFile.setOpenLineCount(matchOpenLineRowsNum.size());

        //保存当前文件的开源信息到mongo库中
        MatchOpenFileMongoDto matchOpenFileMongo = new MatchOpenFileMongoDto();
        matchOpenFileMongo.setId(IdGenerator.uuid32())
                .setFileName(analysisFile.getName())
                .setFilePath(analysisFile.getFileUrl())
                .setOpenType(analysisFile.getOpenType())
                .setFeatureSimilarity(featureSimilarity.floatValue())
                .setOpenRate(openRate.floatValue())
                .setAnalysisType(AnalysisLevelEnum.FUNCTION_LEVEL.getCode())
                .setSubMatchOpenFiles(matchOpenFilesRes);
        mongoTemplate.save(matchOpenFileMongo);
    }


    /**
     * 计算当前文件的特征相似度 和 开源率
     *
     * @param matchOpenFiles           通过MD5 匹配到的所有开源文件
     * @param sourceFileBaseCoreName   当前文件特征文件的 solr coreName
     * @param matchLineRowsNum         所有开源文件匹配到的开源行号列表
     * @param matchFeatureFunctionMd5s 所有开源文件匹配到的特征函数MD5
     *                                 return 匹配的开源文件解析后的结果集
     */
    private List<MatchOpenFile> calculateSimilarityAndOpenRate(SolrDocumentList matchOpenFiles, CodeFile fileAnalysisRes, String sourceFileBaseCoreName, Set<Integer> matchLineRowsNum, Set<String> matchFeatureFunctionMd5s) {

        //匹配的开源文件列表
        List<MatchOpenFile> matchOpenFilesRes = new ArrayList<>();

        //按照函数的特征md5进行分组,getter ,setter等方法的 特征值会重复
        Map<String, List<Function>> featureMd5FunctionMap = fileAnalysisRes.getFunctionList().stream().collect(Collectors.groupingBy(Function::getMd5));

        //首先根据文件的MD5查询开源文件的版本ID,和路径信息
        Set<String> openSourceFileMd5s = matchOpenFiles.stream().map(doc -> (String) doc.get("sourceMd5")).collect(Collectors.toSet());
        Map<String, SolrDocument> md5VersionInfoMap = solrUtils.batchQueryVersionIdFromSourceFileBaseBySourceMd5(sourceFileBaseCoreName, openSourceFileMd5s);

        //根据版本ID查询版本的详细信息
        //todo 这里 查询一个版本的信息 需要检索 两个 solr 库 而且还需要检索 versioinTree 后面需要优化
        Set<String> openSourceFileVersionIds = md5VersionInfoMap.values().stream().map(doc -> (String) doc.get("versionId")).collect(Collectors.toSet());
        List<VersionTree> versionTrees = solrUtils.queryBatchVersionInfoByVersionIds(openSourceFileVersionIds);
        Map<String, VersionTree> versionIdVersionInfoMap = versionTrees.stream().collect(Collectors.toMap(VersionTree::getVersionId, java.util.function.Function.identity()));

        //函数总行数
        BigDecimal totalFunctionLineCount = new BigDecimal(fileAnalysisRes.getFunctionList().stream().mapToInt(Function::getCodeRowNum).sum());


        //被测件文本内容
        String sourcefileContent= FileUtil.readUtf8String(analysisFile.getFileUrl());

        //将文本内容解析成行信息,用于后续文件的开源率计算
        List<String> analysisFileLineInfo = SimilarityUtil.getSplitWords(sourcefileContent);

        for (SolrDocument openSourceFile : matchOpenFiles) {

            //开源文件md5
            String openSourceFileMd5 = openSourceFile.getFieldValue("sourceMd5").toString();

            //解析文件的函数特征值
            List<FunctionInfo> openFileFunctionList = getOpenFileFunctionList(openSourceFile);

            //根据源文件的MD5确定需要查询源码库的序号
            String openSourceCodeCoreIndex = openSourceFileMd5.substring(0, 1) + SolrDBConst.CORE_NAME_SUFFIX_SOURCE_FILE_INFO;

            //获取开源文件的文本信息
            SolrDocument openSourceContent = solrUtils.queryOne(openSourceCodeCoreIndex, "sourceFileMd5:" + openSourceFileMd5, "sourceContent");

            //当前文件匹配特征函数总行数
            int currentFileMatchFeatureLineCount = 0;

            //当前文件所匹配的特征函数MD5
            Set<String> currentFileMatchFeatureFunctionMd5 = new HashSet();

            //遍历函数特征MD5
            for (String funFeatureMd5 : featureMd5FunctionMap.keySet()) {
                List<Function> currentFueatureFunctionList = featureMd5FunctionMap.get(funFeatureMd5);
                //源文件的特征函数列表
                for (FunctionInfo openFunction : openFileFunctionList) {
                    if (funFeatureMd5.equals(openFunction.getTraitFunMd5())) {
                        //每个特征函数 不能多次匹配，影响整体特征相似度
                        //匹配成功后，相同的特征行 一并加上
                        if (!currentFileMatchFeatureFunctionMd5.contains(funFeatureMd5)) {
                            currentFileMatchFeatureFunctionMd5.add(funFeatureMd5);
                            matchFeatureFunctionMd5s.add(funFeatureMd5);
                            currentFileMatchFeatureLineCount += currentFueatureFunctionList.stream().mapToInt(Function::getCodeRowNum).sum();
                        }
                    }
                }
            }

            //当前文件的开源率
            Pair<Float, HashSet<Integer>> openRateAndSaveRowNum = SimilarityUtil.getOpenRateAndSaveRowNum(analysisFileLineInfo, openSourceContent.getFieldValue("sourceContent").toString());
            //将当前文件匹配的行号，存储到缓存中，方便统计整体的开源率
            matchLineRowsNum.addAll(openRateAndSaveRowNum.getValue());

            //统计当前文件的特征相似度
            BigDecimal featureSimilarity = new BigDecimal(currentFileMatchFeatureLineCount).divide(totalFunctionLineCount, 4, RoundingMode.HALF_UP).multiply(new BigDecimal(100)).setScale(2);

            SolrDocument openEntries = md5VersionInfoMap.get(openSourceFileMd5);
            VersionTree versionInfo = versionIdVersionInfoMap.get(openEntries.get("versionId"));
            String openFilePath = (String) md5VersionInfoMap.get(openSourceFileMd5).getFieldValue("fullPath");

            //组装当前开源文件的开源项目信息
            MatchOpenFile matchOpenFileInfo = new MatchOpenFile();
            matchOpenFileInfo.setId(IdGenerator.uuid32())
                    .setFileName(FileUtil.getName(openFilePath))
                    .setPName(versionInfo.getProName())
                    .setPId(versionInfo.getProId())
                    .setVersion(versionInfo.getVersionName())
                    .setVersionId(versionInfo.getVersionId())
                    .setSourceFilePath(openFilePath)
                    .setSourceUrl(versionInfo.getDownUrl())
                    .setLicenseType(versionInfo.getLicenseType())
                    .setFeatureSimilarity(featureSimilarity.floatValue())
                    .setOpenRate(openRateAndSaveRowNum.getKey())
                    .setMd5(openSourceFileMd5);
            matchOpenFilesRes.add(matchOpenFileInfo);
        }
        return matchOpenFilesRes;
    }

    /**
     * 根据 特征值 从特征库中检索 具有特征相似的
     *
     * @param featureCoreName 检索的solr 库名称
     * @param functionList
     * @return
     */
    private SolrDocumentList getFeatureSimilarityFromSolr(String featureCoreName, List<Function> functionList) {
        //获取方法快的特征MD5,cutMD5
        Set<String> featureFunctionMd5List = functionList.stream().map(Function::getMd5).collect(Collectors.toSet());
        Set<String> cutFunctionMd5List = functionList.stream().map(Function::getSourceMd5).collect(Collectors.toSet());
        Set<String> queryMd5List = Stream.concat(featureFunctionMd5List.stream(), cutFunctionMd5List.stream()).collect(Collectors.toSet());
        if (CollUtil.isEmpty(queryMd5List)) {
            log.error("特征为空,无法查询:{}", analysisFile.getName());
            return new SolrDocumentList();
        }
        String queryStr = "fun_hay:(" + StringUtils.join(queryMd5List, " OR ") + ")";
//                        log.info("检索函数特征,coreName:{} ,queryStr:{}", featureCoreName, queryStr);
        return solrUtils.query(featureCoreName, queryStr, "sourceMd5,fun_hay");
    }


    /**
     * 获取当前文件的函数特征值
     *
     * @param matchOpenFile
     * @return
     */
    private List<FunctionInfo> getOpenFileFunctionList(SolrDocument matchOpenFile) {
        try {
            //解析文件的函数特征值
            String lineFeatureMd5s = matchOpenFile.getFieldValue("fun_hay").toString();
            lineFeatureMd5s = lineFeatureMd5s
                    .replace("\\\\\\\"", "")
                    .replace("\\", "")
                    .replace("\"{", "{")
                    .replace("}\"", "}");
            return JSONArray.parseArray(lineFeatureMd5s, FunctionInfo.class);
        } catch (Exception e) {
            log.error("解析文件特征值失败", e);
        }
        return new ArrayList<>();
    }

    /**
     * 将特征值插入到mongo库中
     *
     * @param features         特征集合
     * @param lineDataMongoDto 当前分析任务 ，特征信息存储
     * @param
     */
    @Deprecated
    private void insertFeatureValue(List<LineModel> features, LineDataMongoDto lineDataMongoDto) {
        List<LineModel> batchInsertList = new ArrayList<>();
        if (CollectionUtil.isNotEmpty(features)) {
            //这里的批量插入逻辑可以进行校验
            //每10条存一次,解析的数据量如果过大，可能会超过MongoDB数据限制
            int batchInsertStpe = 10;
            int total = 0;
            for (int i = 0; i < features.size(); i++) {
                LineModel lineModel = features.get(i);
                if (total != batchInsertStpe) {
                    batchInsertList.add(lineModel);
                    total++;
                }
                if (i == features.size() - 1 && total != batchInsertStpe) {
                    total = 0;
                    lineDataMongoDto.setId(IdGenerator.uuid32())
                            .setLineModels(batchInsertList);
                    mongoTemplate.insert(lineDataMongoDto);
                }
                if (total == batchInsertStpe) {
                    total = 0;
                    lineDataMongoDto.setId(IdGenerator.uuid32())
                            .setLineModels(batchInsertList);
                    mongoTemplate.insert(lineDataMongoDto);
                    batchInsertList.clear();
                }
            }
        } else {
            lineDataMongoDto.setId(IdGenerator.uuid32());
            mongoTemplate.insert(lineDataMongoDto);
        }
    }

}
