package com.keyware.composeanalysis.task;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.io.FileUtil;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.google.common.collect.Sets;
import com.keyware.common.constant.enums.AnalysisStatusEnum;
import com.keyware.composeanalysis.constant.FixedValue;
import com.keyware.composeanalysis.constant.MongoDBConst;
import com.keyware.composeanalysis.constant.enums.AnalysisLevelEnum;
import com.keyware.composeanalysis.constant.enums.FileAnalysisStatusEnum;
import com.keyware.composeanalysis.entity.AnalysisTask;
import com.keyware.composeanalysis.mongo.*;
import com.keyware.composeanalysis.service.impl.AnalysisTaskServiceImpl;
import com.keyware.composeanalysis.solr.VersionTree;
import com.keyware.composeanalysis.solr.VersionTreeNode;
import com.keyware.composeanalysis.util.AnalysisLogUtil;
import com.keyware.composeanalysis.util.SolrUtils;
import com.keyware.utils.IdGenerator;
import com.mongodb.client.MongoClient;
import lombok.extern.log4j.Log4j2;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.solr.common.SolrDocument;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.core.query.Update;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;


import static org.springframework.data.mongodb.core.query.Criteria.where;

/**
 * @author liuzongren
 * @date 2024/7/26
 * desc 项目级溯源分析任务，先将所有文件进行项目级匹配，匹配不中的文件在进行细致级别的匹配
 * 项目级匹配前，需要完成文件解压工作
 */
@Log4j2
public class PorjectAnalysisTask {
    private MongoTemplate mongoTemplate;
    private MongoTemplate keyswanDBTemplate;
    private AnalysisTask analysisTask;
    private AnalysisTaskServiceImpl analysisService;
    private SolrUtils solrUtils;

    /**
     * 项目级分析
     *
     * @param mongoClient
     * @param analysisTask
     * @param solrUtils
     * @param analysisService
     */
    public PorjectAnalysisTask(MongoClient mongoClient, AnalysisTask analysisTask, SolrUtils solrUtils, AnalysisTaskServiceImpl analysisService) {
        this.analysisService = analysisService;
        keyswanDBTemplate = new MongoTemplate(mongoClient, MongoDBConst.DB_NAME_KEYSWAN);
        this.mongoTemplate = new MongoTemplate(mongoClient, MongoDBConst.DB_NAME_PREFIX + analysisTask.getId());
        this.analysisTask = analysisTask;
        this.solrUtils = solrUtils;
    }


    public void doAnalysis() {
        try {
            long startTime = System.currentTimeMillis();
            //首先从versionbasedata库中匹配当前被测件
            Boolean matchedPrject = matchByProjectMd5();

            //从versionbase 中整体匹配不到项目信息， 拿项目的所有文件去匹配 solr库的versionTree去检索
            if (!matchedPrject) {
                List<FileDataMongoDto> unMatchedFiles = matchByAllFilesMd5();

                //剩余没有匹配文件，用文件的md5去匹配solr库的versionTree
                if (CollectionUtils.isNotEmpty(unMatchedFiles)) {
                    matchByFileMd5s(unMatchedFiles);
                }
            }
            //todo 如果整体耗时较长,將matchOpenFileInfo存储到数据库的逻辑修改成异步的
            log.info("项目级分析完成,用时:" + (System.currentTimeMillis() - startTime) / 1000 + "s");
        } catch (Exception e) {
            AnalysisLogUtil.insert(mongoTemplate, "成分分析失败:" + e.getStackTrace());
            log.error("项目级分析失败,项目名称:" + analysisTask.getFileName(), e);
            LambdaUpdateWrapper<AnalysisTask> updateWrapper = new LambdaUpdateWrapper<>();
            updateWrapper.eq(AnalysisTask::getId, analysisTask.getId())
                    .set(AnalysisTask::getAnalysisStatus, AnalysisStatusEnum.FAIL_ANALYSIS.getCode());
            analysisService.update(null,updateWrapper);
        }
    }

    //项目整体匹配, 查看整个项目是否开源
    private Boolean matchByProjectMd5() {
        //判断当前项目整体是否开源，去版本库中匹配
        Query versionBaseQuery = new Query(where("MD5").is(analysisTask.getMd5()));
        VersionbasedataMongoDto openSourceProject = keyswanDBTemplate.findOne(versionBaseQuery, VersionbasedataMongoDto.class);
        //如果匹配中了开源项目,则将状态设置为开源，并将结果存储到数据库中
        if (openSourceProject != null) {
            analysisTask.setOpenType(true);
            //将匹配中的开源项目信息存入当前文件开源信息中
            ProjectAssemblyMongoDto projectAssembly = new ProjectAssemblyMongoDto();
            projectAssembly.setId(IdGenerator.uuid32())
                    .setFileCount(analysisTask.getFileCount())
                    .setMatchFileCount(analysisTask.getFileCount())
                    .setProjectId(openSourceProject.getProjectId())
                    .setVersionId(openSourceProject.getVersionId())
                    .setVersionName(openSourceProject.getVersionName())
                    .setSemblance(100.00d)
                    .setOpenSourceUrl(openSourceProject.getDownloadUrl());
            //根据版本信息去查询项目名称和许可证信息
            Query projectBaseQuery = new Query(where("ID").is(openSourceProject.getProjectId()));
            ProjectBaseDataMongoDto projectbasedata = keyswanDBTemplate.findOne(projectBaseQuery, ProjectBaseDataMongoDto.class);
            if (projectbasedata != null) {
                projectAssembly.setProjectName(projectbasedata.getName());
                if (StringUtils.isNotEmpty(projectbasedata.getLicenseType())) {
                    projectAssembly.setLicenseType(Arrays.asList(projectbasedata.getLicenseType()));
                }
            }
            //当前文件开源信息存入数据库中
            mongoTemplate.insert(projectAssembly);

            //更新文件分析的状态
            mongoTemplate.update(FileDataMongoDto.class)
                    .matching(where("isDirectory").is(false))
                    .apply(new Update().set("openType", true)
                            .set("openRate", 100.00d)
                            .set("fileAnalysisStatus", FileAnalysisStatusEnum.ANALYSIS_DONE.getCode()))
                    .all();

            //保存具体开源文件信息
            VersionTree openProjectList = solrUtils.queryVersionTreeByVersionId(openSourceProject.getVersionId());
            Query fileQuery = new Query(where("isDirectory").is(false));
            List<FileDataMongoDto> fileDataMongoDtos = mongoTemplate.find(fileQuery, FileDataMongoDto.class);
            saveProjectOpenInfo(openProjectList, fileDataMongoDtos);
            return true;
        }
        return false;
    }

    //通过文件的md5 去匹配开源项目
    private List<FileDataMongoDto> matchByAllFilesMd5() {
        //检索当前项目的所有文件
        Query query = new Query(where("isDirectory").is(false));
        List<FileDataMongoDto> projectFiles = mongoTemplate.query(FileDataMongoDto.class).matching(query).all();

        //分多次拿所有文件匹配solr库
        Set<String> matchedFileMd5Set =  multipleMatchByAllFilesMd5(projectFiles);

        //统计未匹配的文件
        List<FileDataMongoDto> unMatchedFiles = projectFiles.stream().filter(file -> !matchedFileMd5Set.contains(file.getMd5())).collect(Collectors.toList());
        return unMatchedFiles;
    }

    //通过文件的md5去特征库匹配
    private void matchByFileMd5s(List<FileDataMongoDto> unMatchedFiles) {

        //将文件按照后缀分组,方便查询solr库
        Map<String, List<FileDataMongoDto>> allSuffixFiles = unMatchedFiles.stream().filter(file -> StringUtils.isNotEmpty(file.getSuffix())).collect(Collectors.groupingBy(FileDataMongoDto::getSuffix));

        //统计语言的文件
        List<FileDataMongoDto> otherLanguageFiles = new ArrayList<>();

        //遍历主流32语言
        allSuffixFiles.forEach((suffix, data) -> {
            //根据文件后缀名获取特征库名称
            if (FixedValue.SUFFIX_SOLR_VERSION.containsKey(suffix)) {
                String currentCoreName = FixedValue.SUFFIX_SOLR_VERSION.get(suffix);
                //通过md5去*_SourceFileBase中匹配版本Id
                Set<String> fileMd5s = data.stream().map(FileDataMongoDto::getMd5).collect(Collectors.toSet());
                Map<String, SolrDocument> md5VersionObjMap = solrUtils.batchQueryVersionIdFromSourceFileBaseBySourceMd5(currentCoreName, fileMd5s);
                if (CollectionUtil.isEmpty(md5VersionObjMap)) {
                    return;
                }
                //保存结果数据
                saveMatchOpenFileInfo(md5VersionObjMap, data);
            } else {
                //非主流语言的，没有单独的特征库,统一到默认的特征库进行检索
                otherLanguageFiles.addAll(data);
            }
        });

        //将无后缀的文件 归纳于 处理非32种语言的文件
        List<FileDataMongoDto> noSuffixFiles = unMatchedFiles.stream().parallel().filter(file -> StringUtils.isEmpty(file.getSuffix())).collect(Collectors.toList());
        otherLanguageFiles.addAll(noSuffixFiles);

        if (CollectionUtils.isNotEmpty(otherLanguageFiles)) {
            //非32种语言的会分2种MD5
            //暂时忽略字符流md5的匹配，因为大部分都是一样的
            Set<String> fileMd5s = otherLanguageFiles.stream().map(FileDataMongoDto::getMd5).collect(Collectors.toSet());
            Map<String, SolrDocument> md5VersionIdMap = solrUtils.batchQueryVersionIdFromSourceFileBaseBySourceMd5(MongoDBConst.TABLE_NAME_SOURCE_FILE_BASE, fileMd5s);
            if (md5VersionIdMap == null || md5VersionIdMap.isEmpty()) {
                //如果没有匹配到,直接更新文件分析状态已完成，因为非32种语言的文件，无法进行解析，通过源文件的MD5匹配不到，就匹配不到了，无需进行下一步的匹配
                updateFileAnalysisStatus(fileMd5s,true);
                return;
            }
            saveMatchOpenFileInfo(md5VersionIdMap, otherLanguageFiles);
            //直接更改没有匹配的文件分析状态,因为没有匹配上的文件，无法进行下一步的匹配(下一步的匹配只针对32种主流的语言)
            Set<String> notMatchFileMd5s = Sets.difference(fileMd5s, md5VersionIdMap.keySet());
            updateFileAnalysisStatus(notMatchFileMd5s,false);
        }

    }


    //匹配到开源文件后,补充详细信息然后保存到mongo中
    private void saveMatchOpenFileInfo(Map<String, SolrDocument> md5VersionIdMap, List<FileDataMongoDto> originalFiles) {
        List<MatchOpenFileMongoDto> batchInsertCache = new ArrayList<>();
        //根据版本id查询版本的详细信息
        //todo 这段逻辑如果耗时的话，可以异步处理 补充文件的版本信息
        Set<String> versionIds = md5VersionIdMap.values().stream().map(doc->(String)doc.get("versionId")).collect(Collectors.toSet());
        List<VersionTree> versionInfos = solrUtils.queryBatchVersionInfoByVersionIds(versionIds);
        Map<String, VersionTree> versionTreeMap = versionInfos.stream().collect(Collectors.toMap(VersionTree::getVersionId, Function.identity(), (key1, key2) -> key1));
        Map<String, FileDataMongoDto> fileMd5ObjMap = originalFiles.stream().collect(Collectors.toMap(FileDataMongoDto::getMd5, Function.identity(), (key1, key2) -> key1));
        md5VersionIdMap.forEach((fileMd5, versionObj) -> {
            String versionId = (String) versionObj.get("versionId");
            VersionTree versionInfo = versionTreeMap.get(versionId);
            if (versionInfo == null){
                log.error("根据versionId,未在versionTree中找到版本信息,fileMd5:{},versionId:{}",fileMd5, versionId);
                return;
            }
            FileDataMongoDto fileDataMongoDto = fileMd5ObjMap.get(fileMd5);
            MatchOpenFileMongoDto matchOpenFile = getMatchOpenFile(versionInfo, fileDataMongoDto, versionObj.get("fullPath").toString());
            batchInsertCache.add(matchOpenFile);
        });

        if (CollectionUtils.isNotEmpty(batchInsertCache)) {
            mongoTemplate.insert(batchInsertCache, MatchOpenFileMongoDto.class);
            //更新文件分析的状态
            updateFileAnalysisStatus(md5VersionIdMap.keySet(),true);
        }
    }


    //匹配到开源项目后,保存各个文件的开源信息
    private void saveProjectOpenInfo(VersionTree versionInfo, List<FileDataMongoDto> originalFiles) {
        Map<String, FileDataMongoDto> originalMd5ObjMap = originalFiles.stream().collect(Collectors.toMap(FileDataMongoDto::getMd5, Function.identity(), (key1, key2) -> key1));
        Set<String> matchedMd5s = new HashSet<>();
        List<MatchOpenFileMongoDto> batchInsertCache = new ArrayList<>();
        List<VersionTreeNode> fileInfos = versionInfo.getDirTree();

        fileInfos.forEach(versionTreeNodeObj->{
            String openFileMd5 = versionTreeNodeObj.getSourceFileMd5();
            //看是否和被测件的md5匹配
            if (originalMd5ObjMap.keySet().contains(openFileMd5)) {
                //匹配的文件只保存一次
                if (!matchedMd5s.contains(openFileMd5)) {
                    MatchOpenFileMongoDto matchOpenFile = getMatchOpenFile(versionInfo, originalMd5ObjMap.get(openFileMd5),versionTreeNodeObj.getFullPath());
                    batchInsertCache.add(matchOpenFile);
                    matchedMd5s.add(openFileMd5);
                }
            }
            //分批保存，防止单个项目太大，撑爆内存 或 超过 mongodb最大插入数
            if (batchInsertCache.size() >= 1000) {
                mongoTemplate.insert(batchInsertCache, MatchOpenFileMongoDto.class);
                batchInsertCache.clear();
            }
        });

        if (batchInsertCache.size() != 0) {
            mongoTemplate.insert(batchInsertCache, MatchOpenFileMongoDto.class);
        }

        //更新文件分析的状态
        updateFileAnalysisStatus(matchedMd5s,true);
    }

    //获取匹配到的开源文件信息
    private MatchOpenFileMongoDto getMatchOpenFile(VersionTree versionInfo, FileDataMongoDto originalFile,String openFilePath) {
        //设置匹配文件的信息
        MatchOpenFile matchOpenFile = new MatchOpenFile();
        matchOpenFile.setId(IdGenerator.uuid32())
                .setFileName(FileUtil.getName(openFilePath))
                .setPName(versionInfo.getProName())
                .setPId(versionInfo.getProId())
                .setVersion(versionInfo.getVersionName())
                .setVersionId(versionInfo.getVersionId())
                .setSourceFilePath(openFilePath)
                .setSourceUrl(versionInfo.getDownUrl())
                .setLicenseType(versionInfo.getLicenseType())
                .setMd5(originalFile.getMd5())
                .setFeatureSimilarity(100.00f)
                .setOpenRate(100.00f);

        //创建当前文件与开源代码的匹配信息
        MatchOpenFileMongoDto matchOpenFileInfo = new MatchOpenFileMongoDto();
        matchOpenFileInfo.setId(IdGenerator.uuid32())
                .setFileName(originalFile.getName())
                .setFilePath(originalFile.getFileUrl())
                .setOpenType(true)
                .setFeatureSimilarity(100.00f)
                .setOpenRate(100.00f)
                .setAnalysisType(AnalysisLevelEnum.FILE_LEVEL.getCode())
                .setSubMatchOpenFiles(Arrays.asList(matchOpenFile));
        return matchOpenFileInfo;
    }

    //匹配拿所有文件的md5去versionTree中,需要分多次匹配，单次匹配多个结果集的话，会导致solr响应长时间阻塞
    //多次匹配,每次匹配上一次未匹配种的文件
    //todo 这里需要设置一个阈值，一共匹配多少次，或者当相似度达到多少的时候，停止整体匹配
    //目前默认查询三次
    private Set<String> multipleMatchByAllFilesMd5(List<FileDataMongoDto> projectFiles) {

        //获取被测件所有文件的md5
        Set<String> projectFilesMd5 = projectFiles.stream().map(FileDataMongoDto::getMd5).collect(Collectors.toSet());

        //匹配到的文件md5存入set中
        Set<String> matchedFileMd5Set = new HashSet<>();

        //剩余未匹配中的文件md5
        Set<String> unMatchedFileMd5s = projectFilesMd5;

        //循环匹配5次,进行整体的文件匹配
        for (int i = 0; i < 5; i++){
            //检索versionTree库
            String queryStr = "dirTree:(" + StringUtils.join(unMatchedFileMd5s, " OR ") + ")";
            log.info("versionTree queryStr: " + queryStr);
            long startTime = System.currentTimeMillis();
            VersionTree openProject = solrUtils.queryVersionTree(queryStr);
            log.info("query versionTree cost:{}s", (System.currentTimeMillis() - startTime) / 1000);
            //如果存在没有匹配到开源数据的情况,直接退出循环匹配
            if (openProject == null){
                break;
            }

            //保存匹配的开源文件信息
            saveProjectOpenInfo(openProject, projectFiles);

            //获取开源项目的所有文件md5集合
            List<String> openFilesMd5 = openProject.getDirTree().stream().map(VersionTreeNode::getSourceFileMd5).collect(Collectors.toList());
            //获取被测件和开源项目相同的文件
            Set<String> matchedFiles = unMatchedFileMd5s.stream().filter(item -> openFilesMd5.contains(item)).collect(Collectors.toSet());

            //保存已匹配的文件md5，后续需要统计整体的开源率
            matchedFileMd5Set.addAll(matchedFiles);

            //计算与当前项目的相似度
            BigDecimal semblance = new BigDecimal(matchedFiles.size()).divide(new BigDecimal(projectFilesMd5.size()), 4, RoundingMode.HALF_UP).multiply(new BigDecimal(100));

            //当相似度小于20%,不保存项目级的信息
            if (semblance.compareTo(new BigDecimal(20)) < 0){
                break;
            }

            ProjectAssemblyMongoDto projectAssembly = new ProjectAssemblyMongoDto();
            projectAssembly.setId(IdGenerator.uuid32())
                    .setFileCount(openFilesMd5.size())
                    .setMatchFileCount(matchedFiles.size())
                    .setProjectId(openProject.getProId())
                    .setProjectName(openProject.getProName())
                    .setVersionName(openProject.getVersionName())
                    .setOpenSourceUrl(openProject.getDownUrl())
                    .setSemblance(semblance.doubleValue());
            mongoTemplate.insert(projectAssembly);

            //获取未匹配中的文件md5,更新下次匹配的md5集合
            unMatchedFileMd5s = Sets.difference(unMatchedFileMd5s, matchedFiles);
            //如果没有剩余未匹配文件,退出整体匹配
            if (CollectionUtils.isEmpty(unMatchedFileMd5s) ) {
                break;
            }
        }
        return matchedFileMd5Set;
    }

    //更新文件分析的状态
    private void updateFileAnalysisStatus(Set<String> fileMd5Set, Boolean openType) {
        mongoTemplate.update(FileDataMongoDto.class)
                .matching(where("md5").in(fileMd5Set))
                .apply(new Update().set("openType", openType)
                        .set("openRate", openType ? 100.00f : 0f)
                        .set("fileAnalysisStatus", FileAnalysisStatusEnum.ANALYSIS_DONE.getCode()))
                .all();
    }

}
